/******************************************************************************
*******************************************************************************
*******************************************************************************

    libferris
    Copyright (C) 2001 Ben Martin

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    For more details see the COPYING file in the root directory of this
    distribution.

    $Id: WebPhotos.cpp,v 1.5 2007/04/14 21:30:08 ben Exp $

*******************************************************************************
*******************************************************************************
******************************************************************************/

#include "WebPhotos_private.hh"
#include <FerrisDOM.hh>
#include <FerrisOpenSSL.hh>
#include <Configuration_private.hh>


using namespace std;

#ifdef HAVE_WEBPHOTOS_API

#include <X11/Xlib.h>
#include <Imlib2.h>

#include <Ferris/Enamel.hh>
#include "Ferris/FerrisCurl_private.hh"
#include "Ferris/FerrisDOM.hh"

namespace Ferris
{

    PhotoMetadata::PhotoMetadata( fh_context md )
    {
        m_id = toint(getStrAttr( md, "id", "0" ));
        m_secret = getStrAttr( md, "secret", "" );
        m_server = getStrAttr( md, "server", "" );
        m_title  = getStrAttr( md, "title", "" );
        m_isPrimary = isTrue(getStrAttr( md, "isprimary", "0" ));
    }
    
    PhotoMetadata::~PhotoMetadata()
    {
    }
    

    long
    PhotoMetadata::getID()
    {
        return m_id;
    }
    
    std::string PhotoMetadata::getSecret()
    {
        return m_secret;
    }
    
    std::string PhotoMetadata::getServer()
    {
        return m_server;
    }
    
    std::string PhotoMetadata::getTitle()
    {
        return m_title;
    }
    
    bool
    PhotoMetadata::isPrimary()
    {
        return m_isPrimary;
    }
    

    /********************************************************************************/
    /********************************************************************************/
    /********************************************************************************/
    /********************************************************************************/

    WebPhotos::WebPhotos( const std::string& username )
        :
        m_username( username ),
        m_includeEAIsPresentRegex( 0 ),
        m_includeEAandValueRegex( 0 )
    {
        if( m_username.empty() )
        {
            m_username = WebPhotos::getDefaultUsername();
        }
    }
    
    WebPhotos::~WebPhotos()
    {
    }
    
    
    std::string
    WebPhotos::getImplementationShortName()
    {
        return "flickr";
    }
    
    std::string
    WebPhotos::getRESTURL()
    {
        std::string url = "http://www.flickr.com/services/rest/";
        return url;
    }

    std::string
    WebPhotos::getUploadURL()
    {
        string earl = "http://api.flickr.com/services/upload/?";
        return earl;
    }

    
    std::string
    WebPhotos::getAuthURL()
    {
        return "http://www.flickr.com/services/auth/";
    }
    

    std::string
    WebPhotos::getPostUploadURLPrefix()
    {
        return "http://www.flickr.com/tools/uploader_edit.gne?ids=";
    }
    

    bool
    WebPhotos::haveAPIKey()
    {
        try
        {
            fh_context apikey = Resolve( "~/.ferris/flickr-api-key.txt" );
            fh_context apisec = Resolve( "~/.ferris/flickr-api-secret.txt" );

            return !getStrAttr( apikey, "content", "" ).empty()
                && !getStrAttr( apisec, "content", "" ).empty();
        }
        catch( ... )
        {
        }
        return false;
    }

    std::string
    WebPhotos::getAPIKey()
    {
        return getStrSubCtx( "~/.ferris/flickr-api-key.txt", "" );
    }

    std::string
    WebPhotos::getAPISecret()
    {
        return getStrSubCtx( "~/.ferris/flickr-api-secret.txt", "" );
    }

    
    std::string
    WebPhotos::getToken()
    {
        stringstream ss;
        ss << CFG_WEBPHOTOS_TOKEN_K << "-" << getImplementationShortName() << "-" << m_username;
        LG_WEBPHOTO_D << "getting token key:" << ss.str() << endl;
        string ret = getEDBString( FDB_SECURE, tostr(ss), CFG_WEBPHOTOS_TOKEN_DEFAULT );
        LG_WEBPHOTO_D << "have token:" << ret << endl;
        return ret;
    }

    
    void
    WebPhotos::setToken( const std::string& token )
    {
        stringstream ss;
        ss << CFG_WEBPHOTOS_TOKEN_K << "-" << getImplementationShortName() << "-" << m_username;
        LG_WEBPHOTO_D << "setting token key:" << ss.str() << endl;
        setEDBString( FDB_SECURE, tostr(ss), token );
    }

    
    bool
    WebPhotos::haveToken()
    {
        return !getToken().empty();
    }

    std::string
    WebPhotos::getDefaultUsername()
    {
        stringstream ss;
        ss << CFG_WEBPHOTOS_DEFAULT_USERNAME_KEY << "-" << getImplementationShortName();
        string ret = getEDBString( FDB_SECURE, tostr(ss),
                                   CFG_WEBPHOTOS_DEFAULT_USERNAME_DEFAULT );
        return ret;
    }

    void
    WebPhotos::setDefaultUsername( const std::string& s )
    {
        stringstream ss;
        ss << CFG_WEBPHOTOS_DEFAULT_USERNAME_KEY << "-" << getImplementationShortName();
        setEDBString( FDB_SECURE, tostr(ss), s );
    }


    long
    WebPhotos::getDefaultLargestDimension()
    {
        stringstream ss;
        ss << CFG_WEBPHOTOS_LARGEST_DIM_K << "-" << getImplementationShortName();
        string ret = getEDBString( FDB_GENERAL, tostr(ss), "0" );
        return toint(ret);
    }
    
    void
    WebPhotos::setDefaultLargestDimension( long v )
    {
        stringstream ss;
        ss << CFG_WEBPHOTOS_LARGEST_DIM_K << "-" << getImplementationShortName();
        setEDBString( FDB_GENERAL, tostr(ss), tostr(v) );
    }


    std::string
    WebPhotos::getDefaultIncludeEAIsPresentRegexString()
    {
        stringstream ss;
        ss << CFG_WEBPHOTOS_INCLUDE_EA_IS_PRESENT_REGEX_K << "-" << getImplementationShortName();
        string s = getEDBString( FDB_GENERAL, tostr(ss), "" );
        return s;
    }
    
    
    fh_rex
    WebPhotos::getDefaultIncludeEAIsPresentRegex()
    {
        if( m_includeEAIsPresentRegex )
            return m_includeEAIsPresentRegex;

        string s = getDefaultIncludeEAIsPresentRegexString();
        if( s.empty() )
            return 0;

        m_includeEAIsPresentRegex = toregexh( s );
        return m_includeEAIsPresentRegex;
    }
    
    void
    WebPhotos::setDefaultIncludeEAIsPresentRegex( const std::string& s )
    {
        stringstream ss;
        ss << CFG_WEBPHOTOS_INCLUDE_EA_IS_PRESENT_REGEX_K << "-" << getImplementationShortName();
        setEDBString( FDB_GENERAL, tostr(ss), s );
        
        if( s.empty() )
            m_includeEAIsPresentRegex = 0;
        else
            m_includeEAIsPresentRegex = toregexh( s );
    }

    std::string
    WebPhotos::getDefaultIncludeEAandValueRegexString()
    {
        stringstream ss;
        ss << CFG_WEBPHOTOS_INCLUDE_EA_AND_VALUE_REGEX_K << "-" << getImplementationShortName();
        string s = getEDBString( FDB_GENERAL, tostr(ss), "" );
        return s;
    }
    
    
    fh_rex
    WebPhotos::getDefaultIncludeEAandValueRegex()
    {
        if( m_includeEAandValueRegex )
            return m_includeEAandValueRegex;

        string s = getDefaultIncludeEAandValueRegexString();
        if( s.empty() )
            return 0;

        m_includeEAandValueRegex = toregexh( s );
        return m_includeEAandValueRegex;
    }
    
    void
    WebPhotos::setDefaultIncludeEAandValueRegex( const std::string& s )
    {
//        cerr << "setDefaultIncludeEAandValueRegex() s:" << s << endl;
        stringstream ss;
        ss << CFG_WEBPHOTOS_INCLUDE_EA_AND_VALUE_REGEX_K << "-" << getImplementationShortName();
        setEDBString( FDB_GENERAL, tostr(ss), s );

        if( s.empty() )
            m_includeEAandValueRegex = 0;
        else
            m_includeEAandValueRegex = toregexh( s );
    }
    

    bool
    WebPhotos::isDefaultImageProtectionPublic()
    {
        stringstream ss;
        ss << CFG_WEBPHOTOS_DEFAULT_PROTECTION_PUBLIC_K << "-" << getImplementationShortName();
        string ret = getEDBString( FDB_GENERAL, tostr(ss), "0" );
        return isTrue(ret);
    }
    
    void WebPhotos::setDefaultImageProtectionPublic( bool v )
    {
//        cerr << "setDefaultImageProtectionPublic() v:" << v << endl;
        stringstream ss;
        ss << CFG_WEBPHOTOS_DEFAULT_PROTECTION_PUBLIC_K << "-" << getImplementationShortName();
        setEDBString( FDB_GENERAL, tostr(ss), tostr(v) );
    }
    
    bool WebPhotos::isDefaultImageProtectionFriend()
    {
        stringstream ss;
        ss << CFG_WEBPHOTOS_DEFAULT_PROTECTION_FRIEND_K << "-" << getImplementationShortName();
        string ret = getEDBString( FDB_GENERAL, tostr(ss), "0" );
        return isTrue(ret);
    }
        
    void WebPhotos::setDefaultImageProtectionFriend( bool v )
    {
        stringstream ss;
        ss << CFG_WEBPHOTOS_DEFAULT_PROTECTION_FRIEND_K << "-" << getImplementationShortName();
        setEDBString( FDB_GENERAL, tostr(ss), tostr(v) );
    }
    
        
    bool WebPhotos::isDefaultImageProtectionFamily()
    {
        stringstream ss;
        ss << CFG_WEBPHOTOS_DEFAULT_PROTECTION_FAMILY_K << "-" << getImplementationShortName();
        string ret = getEDBString( FDB_GENERAL, tostr(ss), "0" );
        return isTrue(ret);
    }
            
    void WebPhotos::setDefaultImageProtectionFamily( bool v )
    {
        stringstream ss;
        ss << CFG_WEBPHOTOS_DEFAULT_PROTECTION_FAMILY_K << "-" << getImplementationShortName();
        setEDBString( FDB_GENERAL, tostr(ss), tostr(v) );
    }
    
            
    
    std::string
    WebPhotos::getUserName()
    {
        stringstream ss;
        ss << CFG_WEBPHOTOS_USERNAME_K << "-" << getImplementationShortName();
        string ret = getEDBString( FDB_SECURE, tostr(ss), "" );
        return ret;
    }
    
    std::string
    WebPhotos::getFullName()
    {
        stringstream ss;
        ss << CFG_WEBPHOTOS_FULLNAME_K << "-" << getImplementationShortName();
        string ret = getEDBString( FDB_SECURE, tostr(ss), "" );
        return ret;
    }
    


    void
    WebPhotos::setUserName( const std::string& s )
    {
        m_username = s;
        stringstream ss;
        ss << CFG_WEBPHOTOS_USERNAME_K << "-" << getImplementationShortName();
        setEDBString( FDB_SECURE, tostr(ss), s );
    }
    
    void
    WebPhotos::setFullName( const std::string& s )
    {
        stringstream ss;
        ss << CFG_WEBPHOTOS_FULLNAME_K << "-" << getImplementationShortName();
        setEDBString( FDB_SECURE, tostr(ss), s );
    }
    

    /********************************************************************************/
    /********************************************************************************/
    /********************************************************************************/

    fh_context
    WebPhotos::handleCallIO( stringmap_t args )
    {
        std::string url = getRESTURL();
        
        
        fh_context ret = 0;
        stringstream callss;
        callss << url << "?";
        for( stringmap_t::iterator iter = args.begin(); iter != args.end(); ++iter )
        {
            callss << iter->first << "=" << iter->second << "&";
        }

        string callstring = callss.str();
        LG_WEBPHOTO_W << "call string:" << callstring << endl;

        try
        {
            fh_context c = Resolve( callstring );
            string content = getStrAttr( c, "content", "<error/>", true, true );
            LG_WEBPHOTO_W << "raw result:" << content << endl;

            fh_domdoc doc = Factory::StringToDOM( content );
            DOMElement* root = doc->getDocumentElement();
            std::string statcode = getAttribute( root, "stat" );
            if( statcode != "" && statcode != "ok" )
            {
                stringstream ss;
                ss << "Error. Data from server:" << content << endl;
                LG_WEBPHOTO_W << tostr(ss);
                Throw_WebPhotoException( tostr(ss), 0 );
            }

            ret = Factory::mountDOM( doc );
            if( !ret->isSubContextBound( "rsp" ) )
            {
                stringstream ss;
                ss << "Error. No RSP element at top level of data." << endl
                   << "Data from server:" << tostream( doc, true ) << endl;
                LG_WEBPHOTO_W << tostr(ss);
                Throw_WebPhotoException( tostr(ss), 0 );
            }
            ret = ret->getSubContext( "rsp" );
            doc = 0;
            
            return ret;
        }
        catch( exception& e )
        {
            LG_WEBPHOTO_W << "e:" << e.what() << endl;
            throw;
        }

        return ret;
    }

    std::string
    WebPhotos::makeSignedString( stringmap_t args )
    {
        stringstream ss;
        ss << getAPISecret();
        for( stringmap_t::iterator iter = args.begin(); iter != args.end(); ++iter )
        {
            ss << iter->first << iter->second;
        }
        LG_WEBPHOTO_W << "signed_string() args:" << ss.str() << endl;
        std::string md5string = digest( ss.str(), "md5" );
        return md5string;
    }

    
    fh_context
    WebPhotos::signedCall( std::string methodname,
                           stringmap_t args,
                           bool dontAutoIncludeToken )
    {
        if( !dontAutoIncludeToken )
        {
            string token = getToken();
            args[ "auth_token" ] = token;
            LG_WEBPHOTO_W << "Adding token:" << token << endl;
        }
        args["method"]  = methodname;
        args["api_key"] = getAPIKey();
        LG_WEBPHOTO_W << "methodname:" << methodname << endl;
        
        string md5string = makeSignedString( args );

        stringmap_t callargs = args;
        callargs.insert( make_pair( "api_sig", md5string ) );
        fh_context doc = handleCallIO( callargs );
        return doc;
    }

//     fh_context
//     WebPhotos::SignedCall( std::string methodname, stringmap_t args )
//     {
//         return signedCall( getDefaultUsername(), methodname, args );
//     }

    photolist_t
    WebPhotos::getMyRecent( time_t tt, int pagenum, int limit )
    {
        photolist_t ret;

        tt = Time::getTime() - (30*24*3600);
        
        stringmap_t args;
        args["api_key"] = getAPIKey();
        args["min_date"] = tostr(tt);
        args["per_page" ] = tostr(limit);
        args["page"] = tostr(pagenum);
        args["extras"] = "date_upload,date_taken,original_format,geo,tags,machine_tags";

        fh_context res = signedCall( "flickr.photos.recentlyUpdated", args );

        fh_context c = res;
        if( c->isSubContextBound("photos") )
            c = c->getSubContext("photos");

        Context::iterator ci = c->begin();
        Context::iterator ce = c->end();
        
        LG_WEBPHOTO_D << "getMyRecent() sz:" << distance( res->begin(), res->end() ) << endl;
        for( ; ci!=ce; ++ci )
        {
            LG_WEBPHOTO_D << "Have web photo... id:" << getStrAttr( *ci, "id", "none" ) << endl;
            ret.push_back( new PhotoMetadata( *ci ) );
        }

        return ret;
    }
    
    
    /********************************************************************************/
    /********************************************************************************/
    /********************************************************************************/


    WebPhotoUpload::WebPhotoUpload( fh_webPhotos wf )
        :
        m_wf( wf ),
        m_debugVerbose( false ),
        m_includeEAisPresentRegex( 0 ),
        m_includeEAandValueRegex( 0 ),
        m_publicViewable( false ),
        m_FriendViewable( false ),
        m_FamilyViewable( false ),
        m_DescriptionFromEA( "" ),
        m_largestDimensionExplicit( -1 )
    {
        m_verboseStream = Factory::fcnull();
    }

    WebPhotoUpload::~WebPhotoUpload()
    {
    }

    void
    WebPhotoUpload::setLargestDimensionExplicit( long v )
    {
        m_largestDimensionExplicit = v;
    }


    void WebPhotoUpload::setPublicViewable( bool v )
    {
        m_publicViewable = v;
    }

    void WebPhotoUpload::setFriendViewable( bool v )
    {
        m_FriendViewable = v;
    }

    void WebPhotoUpload::setFamilyViewable( bool v )
    {
        m_FamilyViewable = v;
    }

    void
    WebPhotoUpload::setDescriptionFromEA( const std::string& s )
    {
        m_DescriptionFromEA = s;
    }


    void
    WebPhotoUpload::setIncludeEAisPresentRegex( fh_rex r )
    {
        m_includeEAisPresentRegex = r;
    }

    void
    WebPhotoUpload::setIncludeEAandValueRegex( fh_rex r )
    {
        m_includeEAandValueRegex = r;
    }


    void
    WebPhotoUpload::setVerboseStream( fh_ostream v )
    {
        m_verboseStream = v;
    }

    void
    WebPhotoUpload::setDebugVerbose( bool v )
    {
        m_debugVerbose = v;
    }



    stringlist_t&
    WebPhotoUpload::getUploadedPhotoIDList()
    {
        return m_uploadedPhotoIDList;
    }


    string
    WebPhotoUpload::getTagString( fh_context c )
    {
        fh_rex includeEAisPresentRegex = m_includeEAisPresentRegex;
        fh_rex includeEAandValueRegex  = m_includeEAandValueRegex;
    
        if( !includeEAisPresentRegex )
            includeEAisPresentRegex = m_wf->getDefaultIncludeEAIsPresentRegex();
        if( !includeEAandValueRegex )
            includeEAandValueRegex = m_wf->getDefaultIncludeEAandValueRegex();
    
        typedef AttributeCollection::AttributeNames_t an_t;
        an_t an;
        if( includeEAisPresentRegex || includeEAandValueRegex )
            c->getAttributeNames( an );

        stringstream tagSS;
        if( includeEAisPresentRegex )
        {
            for( an_t::iterator ai = an.begin(); ai!=an.end(); ++ai )
            {
                if( regex_match( *ai, includeEAisPresentRegex ) )
                {
                    if( isTrue( getStrAttr( c, *ai, "0" ) ))
                    {
                        tagSS << *ai << " ";
                    }
                }
            }
        }
        if( includeEAandValueRegex )
        {
            for( an_t::iterator ai = an.begin(); ai!=an.end(); ++ai )
            {
                if( regex_match( *ai, includeEAandValueRegex ) )
                {
                    tagSS << *ai << "=" << getStrAttr( c, *ai, "" ) << " ";
                }
            }
        }

        LG_WEBPHOTO_D << "tags:" << tagSS.str() << endl;
        return tagSS.str();
    }

    fh_context
    WebPhotoUpload::makeScaledImage( fh_context c )
    {
        bool m_antialias = 1;
        long MaxDesiredWidthOrHeight = m_wf->getDefaultLargestDimension();

        if( m_largestDimensionExplicit >= 0 )
            MaxDesiredWidthOrHeight = m_largestDimensionExplicit;

        LG_WEBPHOTO_D << "MaxDesiredWidthOrHeight:" << MaxDesiredWidthOrHeight
                      << " wf->getDefaultLargestDimension():" << m_wf->getDefaultLargestDimension()
                      << " m_largestDimensionExplicit:" << m_largestDimensionExplicit
                      << endl;
        LG_WEBPHOTO_D << "c:" << c->getURL() << endl;

        if( !MaxDesiredWidthOrHeight )
        {
            LG_WEBPHOTO_D << "No scaling needed. returning c:" << c->getURL() << endl;
            return c;
        }
        
        
#ifndef HAVE_IMLIB2
        cerr << "Imlib2 is required to perform image resizing!" << endl;
        cerr << "you must rebuild your libferris :(" << endl;
        return c;
#else


        LG_WEBPHOTO_D << "Scaling image at:" << c->getURL() << endl;
        string dst_url = "/tmp/libferris-webupload-temp-image.jpg";
        string m_imageDataCache;
        fh_context img = c;
        string earl = c->getURL();
        string imlibpath = c->getURL();
        imlib_context_set_anti_alias( m_antialias );
        Imlib_Load_Error ierr;
        if( starts_with( earl, "file:" ) )
            imlibpath = c->getDirPath();
        Imlib_Image image = imlib_load_image_with_error_return( earl.c_str(), &ierr );
        if( !image )
        {
            LG_WEBPHOTO_D << "trying to load image with libferris plugins... URL:" << earl << endl;
            /*
             * Attempt to load the image using the rgba-data plugins of libferris
             */
            int width  = toint( getStrAttr( img, "width",  "-1" ));
            int height = toint( getStrAttr( img, "height", "-1" ));
        
            m_imageDataCache = getStrAttr( img, "rgba-32bpp", "", true, false );
            if( m_imageDataCache.empty() || width == -1 || height == -1 )
            {
                LG_WEBPHOTO_W << "Could not load image at:" << earl << " reason:" << ierr << endl;
                cerr << "Could not load image at:" << earl << " reason:" << ierr << endl;
                return 0;
            }

            image = imlib_create_image_using_data(
                width, height, (unsigned int*)m_imageDataCache.data() );
        }

        imlib_context_set_image( image );
        int w = imlib_image_get_width();
        int h = imlib_image_get_height();
        double ratio = MaxDesiredWidthOrHeight;

        if( w < MaxDesiredWidthOrHeight && h < MaxDesiredWidthOrHeight )
            return c;

        if( w > MaxDesiredWidthOrHeight )
        {
            ratio = MaxDesiredWidthOrHeight;
            ratio /= w;
            w = MaxDesiredWidthOrHeight;
            h = (int)( ratio * h );
        }
        if( h > MaxDesiredWidthOrHeight )
        {
            ratio = MaxDesiredWidthOrHeight;
            ratio /= h;
            w = (int)( ratio * w );
            h = MaxDesiredWidthOrHeight;
        }
        Imlib_Image out_image =
            imlib_create_cropped_scaled_image(0, 0,
                                              imlib_image_get_width(),
                                              imlib_image_get_height(),
                                              w, h);
        imlib_free_image();
        imlib_context_set_image(out_image);
    
    
        imlib_image_set_format( "jpg" );
        imlib_save_image( dst_url.c_str() );
        imlib_free_image();

        fh_context ret = Resolve(  dst_url );
        return ret;
#endif
    }

    void
    WebPhotoUpload::throwIfStatusIsBad( fh_context contextMaybeNull, const std::string& res )
    {
        string earl = "<unknown>";
        if( contextMaybeNull )
            earl = contextMaybeNull->getURL();
        
        fh_domdoc doc = Factory::StringToDOM( res );
        DOMElement* root = doc->getDocumentElement();
        std::string statcode = getAttribute( root, "stat" );
        if( statcode != "" && statcode != "ok" )
        {
            cerr << "processing file:" << earl << endl;
            cerr << "Error. Data from server:" << res << endl;

            stringstream ss;
            ss << "Failed to upload image from:" << earl << endl;
            ss << "raw reason:" << res << endl;
            Throw_WebPhotoException( tostr(ss), 0 );
        }
    }
    
    
    void
    WebPhotoUpload::post( stringmap_t args )
    {
        fh_webPhotos wf = m_wf;
        string earl = wf->getRESTURL();

        FerrisCurl cu;
        m_verboseStream << "Setting DebugVerbose to:" << m_debugVerbose << endl;
        cu.setDebugVerbose( m_debugVerbose );

        args["api_key"] = wf->getAPIKey();
        args["auth_token"] = wf->getToken();
        string sig = wf->makeSignedString( args );
        args["api_sig"] = sig;

        LG_WEBPHOTO_D << "api_key:" << wf->getAPIKey() << endl;
        LG_WEBPHOTO_D << "sig:" << sig << endl;

        cu.setHttpAccept("Content-Type: application/xml");

        m_verboseStream << "Sending POST...to url:" << earl << endl;

        cu.setMultiPartPostData( args );
        string res = cu.post( earl, "" );

        m_verboseStream << "Have result:" << res << endl;
        throwIfStatusIsBad( 0, res );
    }
    
    
    void
    WebPhotoUpload::upload( fh_context c, const std::string& title, const std::string& desc )
    {
        fh_webPhotos wf = m_wf;
        m_verboseStream << "Uploading image from:" << c->getURL()
                        << " explicit title:" << title
                        << " explicit desc:" << desc
                        << endl;
        LG_WEBPHOTO_D << "Uploading image from:" << c->getURL()
                      << " explicit title:" << title
                      << " explicit desc:" << desc
                      << endl;
        LG_WEBPHOTO_D  << "pub:" << m_publicViewable
                       << " fri:" << m_FriendViewable
                       << " fam:" << m_FamilyViewable
                       << endl;
    
        FerrisCurl cu;
        m_verboseStream << "Setting DebugVerbose to:" << m_debugVerbose << endl;
        cu.setDebugVerbose( m_debugVerbose );
                
        stringmap_t args;
        args["api_key"] = wf->getAPIKey();
        args["is_public"] = tostr(m_publicViewable);
        args["is_friend"] = tostr(m_FriendViewable);
        args["is_family"] = tostr(m_FamilyViewable);
        args["auth_token"] = wf->getToken();
        LG_WEBPHOTO_D << "before adding tags, args:"
                      << Util::CreateKeyValueString( args ) << endl;
        args["tags"] = getTagString( c );
        if( !title.empty() )  args["title"] = title;
        if( !desc.empty() )   args["description"] = desc;
        if( !m_DescriptionFromEA.empty() )
        {
            string v = getStrAttr( c, m_DescriptionFromEA, "" );
            if( !v.empty() )
                args["description"] = v;
        }
        string sig = wf->makeSignedString( args );
        args["api_sig"] = sig;
        fh_context scaledc = makeScaledImage( c );
        if( !scaledc )
        {
            stringstream ss;
            ss << "Failed to scale image from:" << c->getURL() << endl;
            Throw_WebPhotoException( tostr(ss), 0 );
        }
        LG_WEBPHOTO_D << "Adding to upload real scaledc:" << scaledc->getURL() << endl;
        cu.appendFileToUpload( scaledc, c->getDirName(), "", "photo" );

        LG_WEBPHOTO_D << "api_key:" << wf->getAPIKey() << endl;
        LG_WEBPHOTO_D << "sig:" << sig << endl;
                
        cu.setHttpAccept("Content-Type: application/xml");
        string earl = wf->getUploadURL();

        m_verboseStream << "Sending POST...to url:" << earl << endl;

        cu.setMultiPartPostData( args );
        string res = cu.post( earl, "" );

        m_verboseStream << "Have result:" << res << endl;

        throwIfStatusIsBad( c, res );
        
        //<rsp stat="ok">
        //  <photoid>438542969</photoid>
        //</rsp>
    
        fh_domdoc doc = Factory::StringToDOM( res );
        DOMElement* root = doc->getDocumentElement();
        if( DOMElement* e = XML::getChildElement( root, "photoid" ) )
        {
            string id = XML::getChildText( e );
            LG_WEBPHOTO_D   << "uploaded id:" << id << endl;
            m_verboseStream << "uploaded id:" << id << endl;
            m_uploadedPhotoIDList.push_back(id);
        }
    }

    string
    WebPhotoUpload::getPostUploadURL()
    {
        cerr << "m_uploadedPhotoIDList.sz:" << m_uploadedPhotoIDList.size() << endl;
        stringstream ss;
        ss << m_wf->getPostUploadURLPrefix();
        ss << Util::createSeperatedList( m_uploadedPhotoIDList );
        return ss.str();
    }
    

    /********************************************************************************/
    /********************************************************************************/
    /********************************************************************************/


    class WebPhotos23hq
        :
        public WebPhotos
    {
    protected:

        virtual std::string getImplementationShortName()
            {
                return "23hq";
            }
        
        virtual std::string getRESTURL()
            {
                std::string url = "http://www.23hq.com/services/rest/";
                return url;
            }
        
    public:

        string getUploadURL()
            {
                string earl = "http://www.23hq.com/services/upload/?";
                return earl;
            }

        string getAuthURL()
            {
                return "http://www.23hq.com/services/auth/";
            }
    
        
        
        virtual std::string getPostUploadURLPrefix()
            {
                return "http://www.23hq.com/tools/uploader_edit.gne?ids=";
            }
        
        virtual std::string getAPIKey()
            {
                return getStrSubCtx( "~/.ferris/23hq-api-key.txt", "" );
            }
        
        virtual std::string getAPISecret()
            {
                return getStrSubCtx( "~/.ferris/23hq-api-secret.txt", "" );
            }
        
        
    public:

        WebPhotos23hq( const std::string& username )
            :
            WebPhotos( username )
            {
            }
        
        virtual ~WebPhotos23hq()
            {}
        
    };



    /********************************************************************************/
    /********************************************************************************/
    /********************************************************************************/


    class WebPhotosZooomr
        :
        public WebPhotos
    {
    protected:

        virtual std::string getImplementationShortName()
            {
                return "zooomr";
            }
        
        virtual std::string getRESTURL()
            {
                std::string url = "http://beta.zooomr.com/bluenote/api/rest/";
                return url;
            }
        
    public:

        string getUploadURL()
            {
                string earl = "http://beta.zooomr.com/bluenote/api/upload/?";
                return earl;
            }

        string getAuthURL()
            {
                return "http://beta.zooomr.com/auth/";
            }
    
        
        
        virtual std::string getPostUploadURLPrefix()
            {
                return "http://beta.zooomr.com/tools/uploader_edit.gne?ids=";
            }
        
        virtual std::string getAPIKey()
            {
                return getStrSubCtx( "~/.ferris/zooomr-api-key.txt", "" );
            }
        
        virtual std::string getAPISecret()
            {
                return getStrSubCtx( "~/.ferris/zooomr-api-secret.txt", "" );
            }
        
        
    public:

        WebPhotosZooomr( const std::string& username )
            :
            WebPhotos( username )
            {
            }
        
        virtual ~WebPhotosZooomr()
            {}
        
    };



    
    

    /********************************************************************************/
    /********************************************************************************/
    /********************************************************************************/
    
    
    
    namespace Factory
    {
        fh_webPhotos getDefaultFlickrWebPhotos()
        {
            static fh_webPhotos ret = 0;
            if( !ret )
            {
                ret = new WebPhotos( "" );
            }
            return ret;
        }

        fh_webPhotos getDefault23hqWebPhotos()
        {
            static fh_webPhotos ret = 0;
            if( !ret )
            {
                ret = new WebPhotos23hq( "" );
            }
            return ret;
        }

        fh_webPhotos getDefaultZooomrWebPhotos()
        {
            static fh_webPhotos ret = 0;
            if( !ret )
            {
                ret = new WebPhotosZooomr( "" );
            }
            return ret;
        }
        
        fh_webPhotos getDefaultWebPhotosForShortName( const std::string& n )
        {
            if( n == "flickr" )
                return getDefaultFlickrWebPhotos();
            if( n == "23hq" )
                return getDefault23hqWebPhotos();
            if( n == "zooomr" )
                return getDefaultZooomrWebPhotos();
            stringstream ss;
            ss << "No default webphoto service by name:" << n << endl;
            Throw_WebPhotoException( tostr(ss), 0 );
        }
        
        
    };
    
    
};
#endif
